/*
 * Copyright © 2018 Red Hat, Inc
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	 See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *       Christian J. Kellner <christian@kellner.me>
 */

#pragma once

#include <glib.h>

G_BEGIN_DECLS

/* how to format uuids */
typedef enum {
  BOLT_UID_FORMAT_FULL  = 0,
  BOLT_UID_FORMAT_SHORT = 1,
  BOLT_UID_FORMAT_ALIAS = 2,
  BOLT_UID_FORMAT_LEN   = 0xFF,

} BoltUuidFormat;

/* generic function  */
char *        bolt_uuid_format (const char *uuid,
                                const char *salt,
                                int         fmt);

int           bolt_uuid_format_from_string (const char *str,
                                            GError    **error);

/*  */
int           format_uid_init (const char *str,
                               GError    **error);

const char *  format_uid (const char *uid);

G_END_DECLS
